/****************************************************************************
 ****************************************************************************
 *
 * codec.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdio.h>

#include "codec.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "ac.h"
#include "picture.h"
#include "transform.h"
#include "codec001.h"
#include "codec002.h"
#include "codec003.h"
#include "codec004.h"
#include "codec005.h"
#include "codec006.h"
#include "codec007.h"



struct functions
	{
	void				(*encode)(struct codec *);
	void				(*decode)(struct codec *);
	};



/****************************************************************************
 * codec_error
 ****************************************************************************/
static void
codec_error(
	struct codec			*cdc)

	{
	error("invalid codec number %d", cdc->number);
	}



/****************************************************************************
 * codec_unknown
 ****************************************************************************/
static const struct functions		codec_unknown[] = { { codec_error, codec_error } };



/****************************************************************************
 * codec_get_functions
 ****************************************************************************/
static const struct functions *
codec_get_functions(
	int				codec)

	{
	static const struct functions	fnc[] =
		{
			{ codec001_encode, codec001_decode },
			{ codec002_encode, codec002_decode },
			{ codec003_encode, codec003_decode },
			{ codec004_encode, codec004_decode },
			{ codec005_encode, codec005_decode },
			{ codec006_encode, codec006_decode },
			{ codec007_encode, codec007_decode }
		};

	if ((codec < 1) || (codec > 7)) return (codec_unknown);
	return (&fnc[codec - 1]);
	}



/****************************************************************************
 * codec_number_ok
 ****************************************************************************/
int
codec_number_ok(
	int				codec)

	{
	if (codec_get_functions(codec) == codec_unknown) return (0);
	return (1);
	}



/****************************************************************************
 * codec_init
 ****************************************************************************/
void
codec_init(
	struct codec			*cdc,
	struct ac			*ac,
	struct transform		*trf,
	int				number,
	int				bitplane,
	int				nr_bitplanes,
	int				pass)

	{
	struct picture			*pic = transform_get_picture(trf);

	*cdc = (struct codec)
		{
		.ac           = ac,
		.cft          = picture_get_coefficient(pic, PICTURE_MODE_PCOEFF),
		.number       = number,
		.xres         = picture_get_xres(pic),
		.yres         = picture_get_yres(pic),
		.level        = transform_get_level(trf),
		.bitplane     = bitplane,
		.nr_bitplanes = nr_bitplanes,
		.pass         = pass
		};
	}



/****************************************************************************
 * codec_encode
 ****************************************************************************/
void
codec_encode(
	struct codec			*cdc)

	{
	codec_get_functions(cdc->number)->encode(cdc);
	ac_flush(cdc->ac);
	}



/****************************************************************************
 * codec_decode
 ****************************************************************************/
void
codec_decode(
	struct codec			*cdc)

	{
	codec_get_functions(cdc->number)->decode(cdc);
	}



/****************************************************************************
 * codec_set_msb
 ****************************************************************************/
void
codec_set_msb(
	struct codec			*cdc,
	int				msb)

	{
	cdc->msb = msb;
	}



/****************************************************************************
 * codec_get_ac
 ****************************************************************************/
struct ac *
codec_get_ac(
	struct codec			*cdc)

	{
	return (cdc->ac);
	}



/****************************************************************************
 * codec_get_msb
 ****************************************************************************/
int
codec_get_msb(
	struct codec			*cdc)

	{
	return (cdc->msb);
	}



/****************************************************************************
 * codec_get_number
 ****************************************************************************/
int
codec_get_number(
	struct codec			*cdc)

	{
	return (cdc->number);
	}
/******************************************************** Karsten Scheibler */
