/****************************************************************************
 ****************************************************************************
 *
 * codec001.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdio.h>

#include "codec001.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "ac.h"
#include "codec.h"
#include "picture.h"



#define CTX_SIGNIF			0
#define CTX_SIGN			1
#define CTX_MAGN			2



/****************************************************************************
 * get_msb
 ****************************************************************************/
static int
get_msb(
	union coefficient		*cft,
	int				size)

	{
	int				i, m, msb = 0;

	for (i = 0; i < size; i++)
		{
		m = pcoeff_get_msb(&cft[i].pcf);
		if (m > msb) msb = m;
		}
	return (msb);
	}



/****************************************************************************
 * block_encode
 ****************************************************************************/
static int
block_encode(
	struct codec			*cdc,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				bitplane)

	{
	struct pcoeff			*pcf;
	int				bit, i, j, x, y;

	for (j = 0, y = ystart; y < yend; y++) for (i = y * cdc->xres + xstart, x = xstart; x < xend; i++, x++)
		{
		pcf = &cdc->cft[i].pcf;
		if (pcoeff_already_finished(pcf, bitplane)) continue;
		j++;
		if (pcoeff_already_significant(pcf, bitplane))
			{
			bit = pcoeff_get_magnitude(pcf, bitplane);
			if (ac_ctx_write_bit(cdc->ac, CTX_MAGN, bit) == -1) return (-1);
			cdc->magn++;
			continue;
			}
		bit = pcoeff_get_significance(pcf, bitplane);
		if (ac_ctx_write_bit(cdc->ac, CTX_SIGNIF, bit) == -1) return (-1);
		cdc->signif++;
		if (! bit) continue;
		bit = pcoeff_get_sign(pcf);
		if (ac_ctx_write_bit(cdc->ac, CTX_SIGN, bit) == -1) return (-1);
		cdc->sign++;
		}
	debug(1, "(%d,%d)-(%d,%d) j = %d", xstart, ystart, xend, yend, j);
	return (0);
	}



/****************************************************************************
 * block_decode
 ****************************************************************************/
static int
block_decode(
	struct codec			*cdc,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				bitplane)

	{
	struct pcoeff			*pcf;
	int				bit, i, j, x, y;

	for (j = 0, y = ystart; y < yend; y++) for (i = y * cdc->xres + xstart, x = xstart; x < xend; i++, x++)
		{
		pcf = &cdc->cft[i].pcf;
		if (pcoeff_already_finished(pcf, bitplane)) continue;
		j++;
		if (pcoeff_already_significant(pcf, bitplane))
			{
			bit = ac_ctx_read_bit(cdc->ac, CTX_MAGN);
			if (bit == -1) return (-1);
			pcoeff_set_magnitude(pcf, bitplane, bit);
			cdc->magn++;
			continue;
			}
		bit = ac_ctx_read_bit(cdc->ac, CTX_SIGNIF);
		if (bit == -1) return (-1);
		cdc->signif++;
		if (! bit) continue;
		bit = ac_ctx_read_bit(cdc->ac, CTX_SIGN);
		if (bit == -1) return (-1);
		pcoeff_set_significance_and_sign(pcf, bitplane, bit);
		cdc->sign++;
		}
	debug(1, "(%d,%d)-(%d,%d) j = %d", xstart, ystart, xend, yend, j);
	return (0);
	}



/****************************************************************************
 * process_bitplane
 ****************************************************************************/
static int
process_bitplane(
	struct codec			*cdc,
	int				(*func)(struct codec *, int, int, int, int, int),
	int				bitplane)

	{
	int				x = cdc->xres >> cdc->level;
	int				y = cdc->yres >> cdc->level;
	int				l;

	if (func(cdc, 0, 0, x, y, bitplane) == -1) return (-1);
	for (l = 0; l < cdc->level; l++, x <<= 1, y <<= 1)
		{
		if (func(cdc, x, 0, x + x,     y, bitplane) == -1) return (-1);
		if (func(cdc, 0, y,     x, y + y, bitplane) == -1) return (-1);
		if (func(cdc, x, y, x + x, y + y, bitplane) == -1) return (-1);
		}
	return (0);
	}



/****************************************************************************
 * iterate
 ****************************************************************************/
static void
iterate(
	struct codec			*cdc,
	int				(*func)(struct codec *, int, int, int, int, int))

	{
	int				b, n, result;

	ac_init2(cdc->ac, 3, 0, 0);
	for (b = cdc->msb, n = 0; (b >= cdc->bitplane) && (n < cdc->nr_bitplanes); b--, n++)
		{
		cdc->signif = cdc->sign = cdc->magn = 0;
		result = process_bitplane(cdc, func, b);
		cdc->sum_signif += cdc->signif;
		cdc->sum_sign   += cdc->sign;
		cdc->sum_magn   += cdc->magn;
		verbose(2, "bitplane %2d: signif =%9d/%-9d, sign =%9d/%-9d, magn =%9d/%-9d",
			b, cdc->signif, cdc->sum_signif, cdc->sign, cdc->sum_sign, cdc->magn, cdc->sum_magn);
		if (result == -1) break;
		}
	}



/****************************************************************************
 * codec001_encode
 ****************************************************************************/
void
codec001_encode(
	struct codec			*cdc)

	{
	cdc->msb = get_msb(cdc->cft, cdc->xres * cdc->yres);
	iterate(cdc, block_encode);
	}



/****************************************************************************
 * codec001_decode
 ****************************************************************************/
void
codec001_decode(
	struct codec			*cdc)

	{
	iterate(cdc, block_decode);
	}
/******************************************************** Karsten Scheibler */
