/****************************************************************************
 ****************************************************************************
 *
 * codec003.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdio.h>

#include "codec003.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "ac.h"
#include "codec.h"
#include "picture.h"



#define TUPEL_SIZE			17

#define SUBBAND_LL			0
#define SUBBAND_LH			1
#define SUBBAND_HL			2
#define SUBBAND_HH			3

#define CTX_BLOCK			0
#define CTX_SIGN			1
#define CTX_MAGN			2
#define CTX_SIGNIF			3
#define CTXS_PER_SUBBAND		5

#define BLOCK_INIT(a, b)		(struct block) { .x = a, .y = b }

struct block
	{
	int				x;
	int				y;
	int				msb;
	};

struct block_params
	{
	struct codec			*cdc;
	struct block			*blk_msb;
	struct block			*blk_signif;
	struct block			*blk_insignif;
	int				xstart;
	int				ystart;
	int				xend;
	int				yend;
	int				xstart2;
	int				ystart2;
	int				xend2;
	int				yend2;
	int				subband;
	int				bitplane;
	};



/****************************************************************************
 * ctx_quantize
 ****************************************************************************/
static int
ctx_quantize(
	int				*tupel)

	{
	int				i, p, n;

	for (i = 0, p = 0; i < 5; i++) p += tupel[i];
	for (i = 5, n = 0; i < TUPEL_SIZE; i++) n += tupel[i];
	if (p + n == 0) return (0);
	if ((p == 1) && (n == 0)) return (1);
	if ((p <= 5) && (n == 0)) return (2);
	if ((p == 0) && (n == 1)) return (3);
	return (4);
	}



/****************************************************************************
 * ctx_get
 ****************************************************************************/
static int
ctx_get(
	struct block_params		*blk_prm,
	int				x,
	int				y)

	{
	if ((x < blk_prm->xstart) || (y < blk_prm->ystart) ||
		(x >= blk_prm->xend) || (y >= blk_prm->yend) ||
		(blk_prm->xstart + blk_prm->ystart == 0)) return (0);
	return (pcoeff_ctx_get1(&blk_prm->cdc->cft[y * blk_prm->cdc->xres + x].pcf));
	}



/****************************************************************************
 * ctx_get2
 ****************************************************************************/
static int
ctx_get2(
	struct block_params		*blk_prm,
	int				x,
	int				y)

	{
	if ((x < blk_prm->xstart2) || (y < blk_prm->ystart2) ||
		(x >= blk_prm->xend2) || (y >= blk_prm->yend2) ||
		(blk_prm->xstart2 + blk_prm->ystart2 == 0)) return (0);
	return (pcoeff_ctx_get2(&blk_prm->cdc->cft[y * blk_prm->cdc->xres + x].pcf));
	}



/****************************************************************************
 * ctx_select
 ****************************************************************************/
static int
ctx_select(
	struct block_params		*blk_prm,
	int				x,
	int				y)

	{
	int				tupel[TUPEL_SIZE];
	int				x2 = x / 2;
	int				y2 = y / 2;

	tupel[0]  = ctx_get2(blk_prm, x2,     y2 - 1);
	tupel[1]  = ctx_get2(blk_prm, x2 - 1, y2    );
	tupel[2]  = ctx_get2(blk_prm, x2    , y2    );
	tupel[3]  = ctx_get2(blk_prm, x2 + 1, y2    );
	tupel[4]  = ctx_get2(blk_prm, x2,     y2 + 1);
	tupel[5]  = ctx_get(blk_prm, x    , y - 1);
	tupel[6]  = ctx_get(blk_prm, x - 1, y    );
	tupel[7]  = ctx_get(blk_prm, x + 1, y    );
	tupel[8]  = ctx_get(blk_prm, x    , y + 1);
	tupel[9]  = ctx_get(blk_prm, x    , y - 2);
	tupel[10] = ctx_get(blk_prm, x - 2, y    );
	tupel[11] = ctx_get(blk_prm, x + 2, y    );
	tupel[12] = ctx_get(blk_prm, x    , y + 2);
	tupel[13] = ctx_get(blk_prm, x - 1, y - 1);
	tupel[14] = ctx_get(blk_prm, x + 1, y - 1);
	tupel[15] = ctx_get(blk_prm, x - 1, y + 1);
	tupel[16] = ctx_get(blk_prm, x + 1, y + 1);
	return (CTX_SIGNIF + CTXS_PER_SUBBAND * blk_prm->subband + ctx_quantize(tupel));
	}



/****************************************************************************
 * set_coords
 ****************************************************************************/
static struct block_params *
set_coords(
	struct block_params		*blk_prm,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				subband)

	{
	blk_prm->xstart  = xstart;
	blk_prm->ystart  = ystart;
	blk_prm->xend    = xend;
	blk_prm->yend    = yend;
	blk_prm->xstart2 = xstart / 2;
	blk_prm->ystart2 = ystart / 2;
	blk_prm->xend2   = xend / 2;
	blk_prm->yend2   = yend / 2;
	blk_prm->subband = subband;
	return (blk_prm);
	}



/****************************************************************************
 * block_too_small
 ****************************************************************************/
static int
block_too_small(
	int				x,
	int				y)

	{
	if (x * y < 256) return (1);
	return (0);
	}



/****************************************************************************
 * block_msb
 ****************************************************************************/
static void
block_msb(
	struct codec			*cdc,
	struct block			*blk,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend)

	{
	int				i, x, y, m;

	*blk = BLOCK_INIT(xstart + xend, ystart + yend);
	for (y = ystart; y < yend; y++) for (i = y * cdc->xres + xstart, x = xstart; x < xend; i++, x++)
		{
		m = pcoeff_get_msb(&cdc->cft[i].pcf);
		if (m > blk->msb) blk->msb = m;
		}
	debug(1, "(%d,%d)-(%d,%d) msb = %d", xstart, ystart, xend, yend, blk->msb);
	}



/****************************************************************************
 * block_get_significance
 ****************************************************************************/
static int
block_get_significance(
	struct block			*blk,
	int				x,
	int				y,
	int				bitplane)

	{
	int				i;

	for (i = 0; blk[i].x + blk[i].y > 0; i++) if ((blk[i].x == x) && (blk[i].y == y)) break;
	debug_error_condition(blk[i].msb > bitplane);
	if (blk[i].msb == bitplane) return (1);
	return (0);
	}



/****************************************************************************
 * block_found
 ****************************************************************************/
static int
block_found(
	struct block			*blk,
	int				x,
	int				y)

	{
	int				i;

	for (i = 0; blk[i].x + blk[i].y > 0; i++) if ((blk[i].x == x) && (blk[i].y == y)) return (1);
	return (0);
	}



/****************************************************************************
 * block_append
 ****************************************************************************/
static void
block_append(
	struct block			*blk,
	int				x,
	int				y)

	{
	int				i;

	for (i = 0; blk[i].x + blk[i].y > 0; i++) ;
	blk[i] = BLOCK_INIT(x, y);
	}



/****************************************************************************
 * block_significance
 ****************************************************************************/
static int
block_significance(
	struct block_params		*blk_prm)

	{
	struct ac			*ac  = blk_prm->cdc->ac;
	struct block			*blk = blk_prm->blk_signif;
	int				x1   = blk_prm->xend - blk_prm->xstart;
	int				y1   = blk_prm->yend - blk_prm->ystart;
	int				x2   = blk_prm->xend + blk_prm->xstart;
	int				y2   = blk_prm->yend + blk_prm->ystart;
	int				bit;

	if (block_too_small(x1, y1)) return (1);
	if (block_found(blk_prm->blk_signif, x2, y2)) return (1);
	if (block_found(blk_prm->blk_insignif, x2, y2)) return (0);
	if (blk_prm->blk_msb != NULL)
		{
		bit = block_get_significance(blk_prm->blk_msb, x2, y2, blk_prm->bitplane);
		if (ac_ctx_write_bit(ac, CTX_BLOCK, bit) == -1) return (-1);
		}
	else
		{
		bit = ac_ctx_read_bit(ac, CTX_BLOCK);
		if (bit == -1) return (-1);
		}
	if (! bit) blk = blk_prm->blk_insignif;
	block_append(blk, x2, y2);
	return (bit);
	}



/****************************************************************************
 * block_update
 ****************************************************************************/
static int
block_update(
	struct block_params		*blk_prm)

	{
	union coefficient		*cft   = blk_prm->cdc->cft;
	int				xres   = blk_prm->cdc->xres;
	int				xstart = blk_prm->xstart;
	int				ystart = blk_prm->ystart;
	int				xend   = blk_prm->xend;
	int				yend   = blk_prm->yend;
	int				i, x, y;

	if (block_significance(blk_prm)) for (y = ystart; y < yend; y++) for (i = y * xres + xstart, x = xstart; x < xend; i++, x++)
		{
		pcoeff_ctx_update(&cft[i].pcf);
		}
	return (0);
	}



/****************************************************************************
 * block_encode
 ****************************************************************************/
static int
block_encode(
	struct block_params		*blk_prm)

	{
	struct pcoeff			*pcf;
	struct ac			*ac      = blk_prm->cdc->ac;
	union coefficient		*cft     = blk_prm->cdc->cft;
	int				xres     = blk_prm->cdc->xres;
	int				xstart   = blk_prm->xstart;
	int				ystart   = blk_prm->ystart;
	int				xend     = blk_prm->xend;
	int				yend     = blk_prm->yend;
	int				bitplane = blk_prm->bitplane;
	int				*signif  = &blk_prm->cdc->signif;
	int				*sign    = &blk_prm->cdc->sign;
	int				*magn    = &blk_prm->cdc->magn;
	int				bit, i, j, x, y;

	/* check block significance */

	bit = block_significance(blk_prm);
	if (bit != 1) return (bit);

	/* process block */

	for (j = 0, y = ystart; y < yend; y++) for (i = y * xres + xstart, x = xstart; x < xend; i++, x++)
		{
		pcf = &cft[i].pcf;
		if (pcoeff_already_finished(pcf, bitplane)) continue;
		j++;
		if (pcoeff_already_significant(pcf, bitplane))
			{
			bit = pcoeff_get_magnitude(pcf, bitplane);
			if (ac_ctx_write_bit(ac, CTX_MAGN, bit) == -1) return (-1);
			pcoeff_ctx_set_magnitude(pcf, bit);
			(*magn)++;
			continue;
			}
		bit = pcoeff_get_significance(pcf, bitplane);
		if (ac_ctx_write_bit(ac, ctx_select(blk_prm, x, y), bit) == -1) return (-1);
		(*signif)++;
		if (! bit) continue;
		pcoeff_ctx_set_significance(pcf);
		bit = pcoeff_get_sign(pcf);
		if (ac_ctx_write_bit(ac, CTX_SIGN, bit) == -1) return (-1);
		(*sign)++;
		}
	debug(1, "(%d,%d)-(%d,%d) j = %d", xstart, ystart, xend, yend, j);
	return (0);
	}



/****************************************************************************
 * block_decode
 ****************************************************************************/
static int
block_decode(
	struct block_params		*blk_prm)

	{
	struct pcoeff			*pcf;
	struct ac			*ac      = blk_prm->cdc->ac;
	union coefficient		*cft     = blk_prm->cdc->cft;
	int				xres     = blk_prm->cdc->xres;
	int				xstart   = blk_prm->xstart;
	int				ystart   = blk_prm->ystart;
	int				xend     = blk_prm->xend;
	int				yend     = blk_prm->yend;
	int				bitplane = blk_prm->bitplane;
	int				*signif  = &blk_prm->cdc->signif;
	int				*sign    = &blk_prm->cdc->sign;
	int				*magn    = &blk_prm->cdc->magn;
	int				bit, i, j, x, y;

	/* check block significance */

	bit = block_significance(blk_prm);
	if (bit != 1) return (bit);

	/* process block */

	for (j = 0, y = ystart; y < yend; y++) for (i = y * xres + xstart, x = xstart; x < xend; i++, x++)
		{
		pcf = &cft[i].pcf;
		if (pcoeff_already_finished(pcf, bitplane)) continue;
		j++;
		if (pcoeff_already_significant(pcf, bitplane))
			{
			bit = ac_ctx_read_bit(ac, CTX_MAGN);
			if (bit == -1) return (-1);
			pcoeff_set_magnitude(pcf, bitplane, bit);
			pcoeff_ctx_set_magnitude(pcf, bit);
			(*magn)++;
			continue;
			}
		bit = ac_ctx_read_bit(ac, ctx_select(blk_prm, x, y));
		if (bit == -1) return (-1);
		(*signif)++;
		if (! bit) continue;
		pcoeff_ctx_set_significance(pcf);
		bit = ac_ctx_read_bit(ac, CTX_SIGN);
		if (bit == -1) return (-1);
		pcoeff_set_significance_and_sign(pcf, bitplane, bit);
		(*sign)++;
		}
	debug(1, "(%d,%d)-(%d,%d) j = %d", xstart, ystart, xend, yend, j);
	return (0);
	}



/****************************************************************************
 * iterate_fractional_bitplane2
 ****************************************************************************/
static int
iterate_fractional_bitplane2(
	struct block_params		*blk_prm,
	int				(*func)(struct block_params *))

	{
	int				x = blk_prm->cdc->xres >> blk_prm->cdc->level;
	int				y = blk_prm->cdc->yres >> blk_prm->cdc->level;
	int				l;

	if (func(set_coords(blk_prm, 0, 0, x, y, SUBBAND_LL)) == -1) return (-1);
	for (l = 0; l < blk_prm->cdc->level; l++, x <<= 1, y <<= 1)
		{
		if (func(set_coords(blk_prm, x, 0, x + x,     y, SUBBAND_LH)) == -1) return (-1);
		if (func(set_coords(blk_prm, 0, y,     x, y + y, SUBBAND_HL)) == -1) return (-1);
		if (func(set_coords(blk_prm, x, y, x + x, y + y, SUBBAND_HH)) == -1) return (-1);
		}
	return (0);
	}



/****************************************************************************
 * iterate_fractional_bitplane
 ****************************************************************************/
static int
iterate_fractional_bitplane(
	struct block_params		*blk_prm,
	int				(*func)(struct block_params *),
	int				pass)

	{
	int				result;

	blk_prm->cdc->signif = 0;
	blk_prm->cdc->sign   = 0;
	blk_prm->cdc->magn   = 0;
	result = iterate_fractional_bitplane2(blk_prm, func);
	blk_prm->cdc->sum_signif += blk_prm->cdc->signif;
	blk_prm->cdc->sum_sign   += blk_prm->cdc->sign;
	blk_prm->cdc->sum_magn   += blk_prm->cdc->magn;
	if (pass > 0) verbose(2, "bitplane %2d: signif =%9d/%-9d, sign =%9d/%-9d, magn =%9d/%-9d",
		blk_prm->bitplane, blk_prm->cdc->signif, blk_prm->cdc->sum_signif,
		blk_prm->cdc->sign, blk_prm->cdc->sum_sign, blk_prm->cdc->magn,
		blk_prm->cdc->sum_magn);
	return (result);
	}



/****************************************************************************
 * iterate_bitplane
 ****************************************************************************/
static int
iterate_bitplane(
	struct block_params		*blk_prm,
	int				(*func)(struct block_params *))

	{
	struct block			blk_insignif[64] = { };

	blk_prm->blk_insignif = blk_insignif;
	if (iterate_fractional_bitplane(blk_prm, func, 1) == -1) return (-1);
	iterate_fractional_bitplane(blk_prm, block_update, -1);
	return (0);
	}



/****************************************************************************
 * iterate
 ****************************************************************************/
static void
iterate(
	struct codec			*cdc,
	struct block			*blk_msb,
	int				(*func)(struct block_params *))

	{
	struct block			blk_signif[64] = { };
	struct block_params		blk_prm = { .cdc = cdc, .blk_msb = blk_msb, .blk_signif = blk_signif };
	int				n;

	ac_init2(cdc->ac, CTX_SIGNIF + 4 * CTXS_PER_SUBBAND, 0, 0);
	for (blk_prm.bitplane = cdc->msb, n = 0; (blk_prm.bitplane >= cdc->bitplane) && (n < cdc->nr_bitplanes); blk_prm.bitplane--, n++)
		{
		if (iterate_bitplane(&blk_prm, func) == -1) break;
		}
	}



/****************************************************************************
 * codec003_encode
 ****************************************************************************/
void
codec003_encode(
	struct codec			*cdc)

	{
	struct block			blk_msb[64] = { };
	int				x = cdc->xres;
	int				y = cdc->yres;
	int				i, l;

	debug_error_condition(cdc->level > 16);
	for (i = l = 0; l < cdc->level; l++)
		{
		x >>= 1;
		y >>= 1;
		block_msb(cdc, &blk_msb[i++], x, 0, x + x,     y);
		block_msb(cdc, &blk_msb[i++], 0, y,     x, y + y);
		block_msb(cdc, &blk_msb[i++], x, y, x + x, y + y);
		if (block_too_small(x >> 1, y >> 1)) break;
		}
	block_msb(cdc, &blk_msb[i], 0, 0, x, y);
	for (cdc->msb = 0; blk_msb[i].x + blk_msb[i].y > 0; i++)
		{
		if (blk_msb[i].msb > cdc->msb) cdc->msb = blk_msb[i].msb;
		}
	iterate(cdc, blk_msb, block_encode);
	}



/****************************************************************************
 * codec003_decode
 ****************************************************************************/
void
codec003_decode(
	struct codec			*cdc)

	{
	debug_error_condition(cdc->level > 16);
	iterate(cdc, NULL, block_decode);
	}
/******************************************************** Karsten Scheibler */
