/****************************************************************************
 ****************************************************************************
 *
 * codec007.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdio.h>
#include <string.h>

#include "codec007.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "ac.h"
#include "codec.h"
#include "picture.h"



#define TUPEL_SIZE			17
#define DISTANCE			5

#define SUBBAND_LL			0
#define SUBBAND_LH			1
#define SUBBAND_HL			2
#define SUBBAND_HH			3

#define CTX_BLOCK			0
#define CTX_MAGN			1
#define CTX_BASE			16
#define CTX_SIGNIF			CTX_BASE
#define CTX_SIGNIF_SELECT1		CTX_BASE
#define CTX_SIGNIF_SELECT2		(CTX_BASE + 7)
#define CTX_SIGNIF_SELECT3		(CTX_BASE + 7 + 10)
#define CTX_SIGNIF_SELECT456		(CTX_BASE + 7 + 10 + 10)
#define CTX_SIGN			(CTX_BASE + 7 + 10 + 10 + 29)
#define CTXS_PER_SUBBAND		(7 + 10 + 10 + 29 + 17)

#define BLOCK_INIT(a, b)		(struct block) { .x = a, .y = b }

struct block
	{
	int				x;
	int				y;
	int				msb;
	};

struct block_params
	{
	struct codec			*cdc;
	struct block			*blk_msb;
	struct block			*blk_signif;
	struct block			*blk_insignif;
	int				(*action)(struct block_params *);
	int				(*select2x2)(struct block_params *);
	int				(*select)(struct block_params *);
	unsigned char			*signif2x2;
	int				xstart;
	int				ystart;
	int				xend;
	int				yend;
	int				xstart2;
	int				ystart2;
	int				xend2;
	int				yend2;
	int				subband;
	int				bitplane;
	int				pass;
	int				ignore_distance;
	int				x;
	int				y;
	int				offset;
	int				x2;
	int				y2;
	int				offset2;
	int				*tupel;
	};



/****************************************************************************
 * ctx_quantize
 ****************************************************************************/
static int
ctx_quantize(
	struct block_params		*blk_prm,
	int				dist)

	{
	int				*tupel = blk_prm->tupel;
	int				i, p, n, ctx = 0;

	for (i = 0, p = 0; i < 5; i++) p += tupel[i];
	for (i = 5, n = 0; i < TUPEL_SIZE; i++) n += tupel[i];
	if (p + n == 0) return (dist);
	if ((p == 1) && (n == 0)) return (3);
	if ((p <= 5) && (n == 0)) return (4);
	if ((p == 0) && (n == 1)) return (5);
	if (tupel[5] >= 1) ctx += 3;
	if (tupel[8] >= 1) ctx += 3;
	if (tupel[6] >= 1) ctx++;
	if (tupel[7] >= 1) ctx++;
	if (ctx >= 5) ctx += 8;
	else if (ctx >= 4) ctx += 6;
	else if (ctx >= 2) ctx += 4;
	else if (ctx >= 1) ctx += 2;
	if ((ctx == 0) || (ctx == 3) || (ctx == 7) || (ctx == 10))
		{
		if (tupel[13] + tupel[15] >= 1) ctx++;
		if (tupel[14] + tupel[16] >= 1) ctx++;
		}
	if (ctx <= 1)
		{
		if (ctx == 1) ctx += 2;
		if (tupel[2] >= 4) ctx++;
		if (tupel[2] >= 2) ctx++;
		}
	else ctx += 4;
	if (ctx == 0)
		{
		if (tupel[9]  + tupel[12] >= 1) ctx++;
		if (tupel[10] + tupel[11] >= 1) ctx++;
		}
	else ctx += 2;
	return (ctx + 6);
	}



/****************************************************************************
 * ctx_get
 ****************************************************************************/
static int
ctx_get(
	struct block_params		*blk_prm,
	int				x,
	int				y)

	{
	if ((x < blk_prm->xstart) || (y < blk_prm->ystart) ||
		(x >= blk_prm->xend) || (y >= blk_prm->yend) ||
		(blk_prm->xstart + blk_prm->ystart == 0)) return (0);
	return (pcoeff_ctx_get1(&blk_prm->cdc->cft[y * blk_prm->cdc->xres + x].pcf));
	}



/****************************************************************************
 * ctx_get2
 ****************************************************************************/
static int
ctx_get2(
	struct block_params		*blk_prm,
	int				x,
	int				y)

	{
	if ((x < blk_prm->xstart2) || (y < blk_prm->ystart2) ||
		(x >= blk_prm->xend2) || (y >= blk_prm->yend2) ||
		(blk_prm->xstart2 + blk_prm->ystart2 == 0)) return (0);
	return (pcoeff_ctx_get2(&blk_prm->cdc->cft[y * blk_prm->cdc->xres + x].pcf));
	}



/****************************************************************************
 * ctx_get_sign
 ****************************************************************************/
static int
ctx_get_sign(
	struct block_params		*blk_prm,
	int				x,
	int				y)

	{
	struct pcoeff			*pcf; 

	if ((x < blk_prm->xstart) || (y < blk_prm->ystart) ||
		(x >= blk_prm->xend) || (y >= blk_prm->yend)) return (0);
	pcf = &blk_prm->cdc->cft[y * blk_prm->cdc->xres + x].pcf;
	if (pcoeff_ctx_get2(pcf) == 0) return (0);
	if (pcoeff_get_sign(pcf)) return (-1);
	return (1);
	}



/****************************************************************************
 * ctx_update_tupel
 ****************************************************************************/
static struct block_params *
ctx_update_tupel(
	struct block_params		*blk_prm)

	{
	int				*tupel = blk_prm->tupel;
	int				x      = blk_prm->x;
	int				y      = blk_prm->y;

	tupel[5]  = ctx_get(blk_prm, x    , y - 1);
	tupel[6]  = ctx_get(blk_prm, x - 1, y    );
	tupel[7]  = ctx_get(blk_prm, x + 1, y    );
	tupel[8]  = ctx_get(blk_prm, x    , y + 1);
	tupel[9]  = ctx_get(blk_prm, x    , y - 2);
	tupel[10] = ctx_get(blk_prm, x - 2, y    );
	tupel[11] = ctx_get(blk_prm, x + 2, y    );
	tupel[12] = ctx_get(blk_prm, x    , y + 2);
	tupel[13] = ctx_get(blk_prm, x - 1, y - 1);
	tupel[14] = ctx_get(blk_prm, x + 1, y - 1);
	tupel[15] = ctx_get(blk_prm, x - 1, y + 1);
	tupel[16] = ctx_get(blk_prm, x + 1, y + 1);
	return (blk_prm);
	}



/****************************************************************************
 * ctx_sign
 ****************************************************************************/
static int
ctx_sign(
	struct block_params		*blk_prm)

	{
	int				x = blk_prm->x;
	int				y = blk_prm->y;
	int				h, v, ctx;

	h = ctx_get_sign(blk_prm, x - 1, y    ) + ctx_get_sign(blk_prm, x + 1, y    );
	v = ctx_get_sign(blk_prm, x,     y - 1) + ctx_get_sign(blk_prm, x,     y + 1);
	if (h < 0) h = -1;
	if (h > 0) h = 1;
	if (v < 0) v = -1;
	if (v > 0) v = 1;
	ctx = 3 * (h + 1) + v + 1;
	if (ctx == 4)
		{
		h = ctx_get_sign(blk_prm, x - 2, y    ) + ctx_get_sign(blk_prm, x + 2, y    );
		v = ctx_get_sign(blk_prm, x,     y - 2) + ctx_get_sign(blk_prm, x,     y + 2);
		if (h < 0) h = -1;
		if (h > 0) h = 1;
		if (v < 0) v = -1;
		if (v > 0) v = 1;
		ctx = 8 + 3 * (h + 1) + v + 1;
		}
	else if (ctx > 4) ctx--;
	return (ctx);
	}



/****************************************************************************
 * ctx_magn
 ****************************************************************************/
static int
ctx_magn(
	struct block_params		*blk_prm)

	{
	struct pcoeff			*pcf = &blk_prm->cdc->cft[blk_prm->offset].pcf;
	int				x    = blk_prm->x;
	int				y    = blk_prm->y;
	int				ctx, s;

	ctx = pcoeff_get_msb(pcf) - blk_prm->bitplane - 1;
	if (ctx >= 3) return (14);
	s  = ctx_get(blk_prm, x    , y - 1);
	s += ctx_get(blk_prm, x - 1, y    );
	s += ctx_get(blk_prm, x + 1, y    );
	s += ctx_get(blk_prm, x    , y + 1);
	if (ctx <= 1)
		{
		if (ctx == 1) ctx = 6;
		if (s >= 1) ctx++;
		if (s >= 2) ctx++;
		if (s >= 4) ctx++;
		if (s >= 8) ctx++;
		if (s >= 12) ctx++;
		}
	else
		{
		if (s >= 16) ctx = 13;
		else ctx = 12;
		}
	return (ctx);
	}



/****************************************************************************
 * ctx_select1
 ****************************************************************************/
static int
ctx_select1(
	struct block_params		*blk_prm)

	{
	int				*tupel = blk_prm->tupel;
	int				x      = blk_prm->x;
	int				y      = blk_prm->y;
	int				ctx    = 0;

	tupel[5] = ctx_get(blk_prm, x    , y - 1);
	tupel[6] = ctx_get(blk_prm, x - 1, y    );
	tupel[7] = ctx_get(blk_prm, x + 1, y    );
	tupel[8] = ctx_get(blk_prm, x    , y + 1);
	if (tupel[5] >= 2) ctx += 3;
	if (tupel[8] >= 2) ctx += 3;
	if (tupel[6] >= 2) ctx++;
	if (tupel[7] >= 2) ctx++;
	if (ctx <= 3) return (-1);
	tupel[13] = ctx_get(blk_prm, x - 1, y - 1);
	tupel[14] = ctx_get(blk_prm, x + 1, y - 1);
	tupel[15] = ctx_get(blk_prm, x - 1, y + 1);
	tupel[16] = ctx_get(blk_prm, x + 1, y + 1);
	ctx -= 4;
	if (ctx == 0)
		{
		if ((tupel[13] >= 2) || (tupel[15] >= 2)) ctx++;
		if ((tupel[14] >= 2) || (tupel[16] >= 2)) ctx++;
		}
	else ctx += 2;
	return (CTX_SIGNIF_SELECT1 + CTXS_PER_SUBBAND * blk_prm->subband + ctx);
	}



/****************************************************************************
 * ctx_select2
 ****************************************************************************/
static int
ctx_select2(
	struct block_params		*blk_prm)

	{
	int				*tupel = blk_prm->tupel;
	int				x      = blk_prm->x;
	int				y      = blk_prm->y;
	int				ctx    = 0;

	tupel[5] = ctx_get(blk_prm, x    , y - 1);
	tupel[6] = ctx_get(blk_prm, x - 1, y    );
	tupel[7] = ctx_get(blk_prm, x + 1, y    );
	tupel[8] = ctx_get(blk_prm, x    , y + 1);
	if (tupel[5] >= 1) ctx += 3;
	if (tupel[8] >= 1) ctx += 3;
	if (tupel[6] >= 1) ctx++;
	if (tupel[7] >= 1) ctx++;
	if (ctx <= 2) return (-1);
	tupel[13] = ctx_get(blk_prm, x - 1, y - 1);
	tupel[14] = ctx_get(blk_prm, x + 1, y - 1);
	tupel[15] = ctx_get(blk_prm, x - 1, y + 1);
	tupel[16] = ctx_get(blk_prm, x + 1, y + 1);
	ctx -= 3;
	if (ctx <= 1)
		{
		if (ctx == 1) ctx += 2;
		if (tupel[13] + tupel[15] >= 1) ctx++;
		if (tupel[14] + tupel[16] >= 1) ctx++;
		}
	else ctx += 4;
	return (CTX_SIGNIF_SELECT2 + CTXS_PER_SUBBAND * blk_prm->subband + ctx);
	}



/****************************************************************************
 * ctx_select3
 ****************************************************************************/
static int
ctx_select3(
	struct block_params		*blk_prm)

	{
	int				*tupel = blk_prm->tupel;
	int				x      = blk_prm->x;
	int				y      = blk_prm->y;
	int				ctx    = 0;

	tupel[5] = ctx_get(blk_prm, x    , y - 1);
	tupel[6] = ctx_get(blk_prm, x - 1, y    );
	tupel[7] = ctx_get(blk_prm, x + 1, y    );
	tupel[8] = ctx_get(blk_prm, x    , y + 1);
	if (tupel[5] >= 1) ctx += 3;
	if (tupel[8] >= 1) ctx += 3;
	if (tupel[6] >= 1) ctx++;
	if (tupel[7] >= 1) ctx++;
	if (ctx == 0) return (-1);
	tupel[13] = ctx_get(blk_prm, x - 1, y - 1);
	tupel[14] = ctx_get(blk_prm, x + 1, y - 1);
	tupel[15] = ctx_get(blk_prm, x - 1, y + 1);
	tupel[16] = ctx_get(blk_prm, x + 1, y + 1);
	ctx--;
	if (ctx == 0)
		{
		if (tupel[13] + tupel[15] >= 1) ctx++;
		if (tupel[14] + tupel[16] >= 1) ctx++;
		}
	else ctx += 2;
	return (CTX_SIGNIF_SELECT3 + CTXS_PER_SUBBAND * blk_prm->subband + ctx);
	}



/****************************************************************************
 * ctx_select4
 ****************************************************************************/
static int
ctx_select4(
	struct block_params		*blk_prm)

	{
	int				ctx;

	ctx = ctx_quantize(ctx_update_tupel(blk_prm), 0);
	if (ctx <= 7) return (-1);
	return (CTX_SIGNIF_SELECT456 + CTXS_PER_SUBBAND * blk_prm->subband + ctx);
	}



/****************************************************************************
 * ctx_select5
 ****************************************************************************/
static int
ctx_select5(
	struct block_params		*blk_prm)

	{
	int				ctx;

	ctx = ctx_quantize(ctx_update_tupel(blk_prm), 0);
	if (ctx <= 2) return (-1);
	return (CTX_SIGNIF_SELECT456 + CTXS_PER_SUBBAND * blk_prm->subband + ctx);
	}



/****************************************************************************
 * ctx_select6
 ****************************************************************************/
static int
ctx_select6(
	struct block_params		*blk_prm)

	{
	int				ctx;

	ctx = ctx_quantize(ctx_update_tupel(blk_prm), pcoeff_dist_get(&blk_prm->cdc->cft[blk_prm->offset].pcf));
	return (CTX_SIGNIF_SELECT456 + CTXS_PER_SUBBAND * blk_prm->subband + ctx);
	}



/****************************************************************************
 * set_distance
 ****************************************************************************/
static void
set_distance(
	struct block_params		*blk_prm)

	{
	union coefficient		*cft   = blk_prm->cdc->cft;
	int				xres   = blk_prm->cdc->xres;
	int				xcent  = blk_prm->x;
	int				ycent  = blk_prm->y;
	int				xstart = xcent - DISTANCE;
	int				ystart = ycent - DISTANCE;
	int				xend   = xcent + DISTANCE + 1;
	int				yend   = ycent + DISTANCE + 1;
	int				d, i, x, y;

	if (blk_prm->ignore_distance) return;
	if (xstart < blk_prm->xstart) xstart = blk_prm->xstart;
	if (ystart < blk_prm->ystart) ystart = blk_prm->ystart;
	if (xend   > blk_prm->xend)   xend   = blk_prm->xend;
	if (yend   > blk_prm->yend)   yend   = blk_prm->yend;
	for (y = ystart; y < yend; y++) for (i = y * xres + xstart, x = xstart; x < xend; i++, x++)
		{
		d = 2;
		if ((x < xcent - 2) || (x > xcent + 2)) d = 1;
		if ((y < ycent - 2) || (y > ycent + 2)) d = 1;
		pcoeff_dist_set(&cft[i].pcf, d);
		}
	}



/****************************************************************************
 * set_coords
 ****************************************************************************/
static struct block_params *
set_coords(
	struct block_params		*blk_prm,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				subband)

	{
	blk_prm->xstart  = xstart;
	blk_prm->ystart  = ystart;
	blk_prm->xend    = xend;
	blk_prm->yend    = yend;
	blk_prm->xstart2 = xstart / 2;
	blk_prm->ystart2 = ystart / 2;
	blk_prm->xend2   = xend / 2;
	blk_prm->yend2   = yend / 2;
	blk_prm->subband = subband;
	return (blk_prm);
	}



/****************************************************************************
 * set_pass
 ****************************************************************************/
static struct block_params *
set_pass(
	struct block_params		*blk_prm,
	int				(*action)(struct block_params *),
	int				(*select2x2)(struct block_params *),
	int				(*select)(struct block_params *),
	int				pass)

	{
	blk_prm->action    = action;
	blk_prm->select2x2 = select2x2;
	blk_prm->select    = select;
	blk_prm->pass      = pass;
	return (blk_prm);
	}



/****************************************************************************
 * select2x2_significance1
 ****************************************************************************/
static int
select2x2_significance1(
	struct block_params		*blk_prm)

	{
	unsigned char			*signif2x2 = blk_prm->signif2x2;
	int				xres2      = blk_prm->cdc->xres / 2;
	int				offset2    = blk_prm->offset2;
	int				s          = signif2x2[offset2];

	if (s == 4) return (0);
	if (blk_prm->y2 > 0) s += signif2x2[offset2 - xres2];
	if (blk_prm->x2 > 0) s += signif2x2[offset2 - 1];
	if (blk_prm->y2 + 1 < blk_prm->yend2) s += signif2x2[offset2 + xres2];
	if (blk_prm->x2 + 1 < blk_prm->xend2) s += signif2x2[offset2 + 1];
	return ((s != 0) ? 1 : 0);
	}



/****************************************************************************
 * select2x2_significance2
 ****************************************************************************/
static int
select2x2_significance2(
	struct block_params		*blk_prm)

	{
	unsigned char			*signif2x2 = blk_prm->signif2x2;
	int				*tupel     = blk_prm->tupel;
	int				xres2      = blk_prm->cdc->xres / 2;
	int				xend2      = blk_prm->xend2;
	int				yend2      = blk_prm->yend2;
	int				x2         = blk_prm->x2;
	int				y2         = blk_prm->y2;
	int				offset2    = blk_prm->offset2;
	int				s          = signif2x2[offset2];

	if (s == 4) return (0);
	tupel[0] = ctx_get2(blk_prm, x2,     y2 - 1);
	tupel[1] = ctx_get2(blk_prm, x2 - 1, y2    );
	tupel[2] = ctx_get2(blk_prm, x2    , y2    );
	tupel[3] = ctx_get2(blk_prm, x2 + 1, y2    );
	tupel[4] = ctx_get2(blk_prm, x2,     y2 + 1);
	if (y2 > 0) s += signif2x2[offset2 - xres2];
	if (x2 > 0) s += signif2x2[offset2 - 1];
	if (y2 + 1 < yend2) s += signif2x2[offset2 + xres2];
	if (x2 + 1 < xend2) s += signif2x2[offset2 + 1];
	if ((y2 > 0) && (x2 > 0)) s += signif2x2[offset2 - xres2 - 1];
	if ((y2 > 0) && (x2 + 1 < xend2)) s += signif2x2[offset2 - xres2 + 1];
	if ((y2 + 1 < yend2) && (x2 > 0)) s += signif2x2[offset2 + xres2 - 1];
	if ((y2 + 1 < yend2) && (x2 + 1 < xend2)) s += signif2x2[offset2 + xres2 + 1];
	s += tupel[0] + tupel[1] + tupel[2] + tupel[3] + tupel[4];
	return ((s != 0) ? 1 : 0);
	}



/****************************************************************************
 * select2x2_significance3
 ****************************************************************************/
static int
select2x2_significance3(
	struct block_params		*blk_prm)

	{
	int				*tupel = blk_prm->tupel;
	int				x2     = blk_prm->x2;
	int				y2     = blk_prm->y2;

	if (blk_prm->signif2x2[blk_prm->offset2] == 4) return (0);
	tupel[0] = ctx_get2(blk_prm, x2,     y2 - 1);
	tupel[1] = ctx_get2(blk_prm, x2 - 1, y2    );
	tupel[2] = ctx_get2(blk_prm, x2    , y2    );
	tupel[3] = ctx_get2(blk_prm, x2 + 1, y2    );
	tupel[4] = ctx_get2(blk_prm, x2,     y2 + 1);
	return (1);
	}



/****************************************************************************
 * select2x2_magnitude
 ****************************************************************************/
static int
select2x2_magnitude(
	struct block_params		*blk_prm)

	{
	return ((blk_prm->signif2x2[blk_prm->offset2] != 0) ? 1 : 0);
	}



/****************************************************************************
 * select2x2_update
 ****************************************************************************/
static int
select2x2_update(
	struct block_params		*blk_prm)

	{
	return (1);
	}



/****************************************************************************
 * action_encode_significance
 ****************************************************************************/
static int
action_encode_significance(
	struct block_params		*blk_prm)

	{
	struct pcoeff			*pcf     = &blk_prm->cdc->cft[blk_prm->offset].pcf;
	int				bitplane = blk_prm->bitplane;
	int				bit, ctx;

	if (pcoeff_already_finished(pcf, bitplane)) return (0);
	if (pcoeff_already_significant(pcf, bitplane)) return (0);
	if (pcoeff_ctx_already_done(pcf)) return (0);
	ctx = blk_prm->select(blk_prm);
	if (ctx == -1) return (0);
	bit = pcoeff_get_significance(pcf, bitplane);
	if (ac_ctx_write_bit(blk_prm->cdc->ac, ctx, bit) == -1) return (-1);
	blk_prm->cdc->signif++;
	pcoeff_ctx_set_done(pcf);
	if (! bit) return (0);
	pcoeff_ctx_set_significance(pcf);
	blk_prm->signif2x2[blk_prm->offset2]++;
	ctx = CTX_SIGN + blk_prm->subband * CTXS_PER_SUBBAND + ctx_sign(blk_prm);
	bit = pcoeff_get_sign(pcf);
	if (ac_ctx_write_bit(blk_prm->cdc->ac, ctx, bit) == -1) return (-1);
	set_distance(blk_prm);
	blk_prm->cdc->sign++;
	return (0);
	}



/****************************************************************************
 * action_encode_magnitude
 ****************************************************************************/
static int
action_encode_magnitude(
	struct block_params		*blk_prm)

	{
	struct pcoeff			*pcf     = &blk_prm->cdc->cft[blk_prm->offset].pcf;
	int				bitplane = blk_prm->bitplane;
	int				bit;

	if (pcoeff_already_finished(pcf, bitplane)) return (0);
	if (! pcoeff_already_significant(pcf, bitplane)) return (0);
	bit = pcoeff_get_magnitude(pcf, bitplane);
	if (ac_ctx_write_bit(blk_prm->cdc->ac, CTX_MAGN + ctx_magn(blk_prm), bit) == -1) return (-1);
	pcoeff_ctx_set_magnitude(pcf, bit);
	blk_prm->cdc->magn++;
	return (0);
	}



/****************************************************************************
 * action_decode_significance
 ****************************************************************************/
static int
action_decode_significance(
	struct block_params		*blk_prm)

	{
	struct pcoeff			*pcf     = &blk_prm->cdc->cft[blk_prm->offset].pcf;
	int				bitplane = blk_prm->bitplane;
	int				bit, ctx;

	if (pcoeff_already_finished(pcf, bitplane)) return (0);
	if (pcoeff_already_significant(pcf, bitplane)) return (0);
	if (pcoeff_ctx_already_done(pcf)) return (0);
	ctx = blk_prm->select(blk_prm);
	if (ctx == -1) return (0);
	bit = ac_ctx_read_bit(blk_prm->cdc->ac, ctx);
	if (bit == -1) return (-1);
	blk_prm->cdc->signif++;
	pcoeff_ctx_set_done(pcf);
	if (! bit) return (0);
	pcoeff_ctx_set_significance(pcf);
	blk_prm->signif2x2[blk_prm->offset2]++;
	ctx = CTX_SIGN + blk_prm->subband * CTXS_PER_SUBBAND + ctx_sign(blk_prm);
	bit = ac_ctx_read_bit(blk_prm->cdc->ac, ctx);
	if (bit == -1) return (-1);
	pcoeff_set_significance_and_sign(pcf, bitplane, bit);
	set_distance(blk_prm);
	blk_prm->cdc->sign++;
	return (0);
	}



/****************************************************************************
 * action_decode_magnitude
 ****************************************************************************/
static int
action_decode_magnitude(
	struct block_params		*blk_prm)

	{
	struct pcoeff			*pcf     = &blk_prm->cdc->cft[blk_prm->offset].pcf;
	int				bitplane = blk_prm->bitplane;
	int				bit;

	if (pcoeff_already_finished(pcf, bitplane)) return (0);
	if (! pcoeff_already_significant(pcf, bitplane)) return (0);
	bit = ac_ctx_read_bit(blk_prm->cdc->ac, CTX_MAGN + ctx_magn(blk_prm));
	if (bit == -1) return (-1);
	pcoeff_set_magnitude(pcf, bitplane, bit);
	pcoeff_ctx_set_magnitude(pcf, bit);
	blk_prm->cdc->magn++;
	return (0);
	}



/****************************************************************************
 * action_update
 ****************************************************************************/
static int
action_update(
	struct block_params		*blk_prm)

	{
	pcoeff_ctx_update(&blk_prm->cdc->cft[blk_prm->offset].pcf);
	return (0);
	}



/****************************************************************************
 * action2x2
 ****************************************************************************/
static int
action2x2(
	struct block_params		*blk_prm)

	{
	if (! blk_prm->select2x2(blk_prm)) return (0);

	blk_prm->x = 2 * blk_prm->x2;
	blk_prm->y = 2 * blk_prm->y2;
	blk_prm->offset = blk_prm->y * blk_prm->cdc->xres + blk_prm->x;
	if (blk_prm->action(blk_prm) == -1) return (-1);

	blk_prm->x++;
	blk_prm->offset++;
	if (blk_prm->action(blk_prm) == -1) return (-1);

	blk_prm->x--;
	blk_prm->y++;
	blk_prm->offset += blk_prm->cdc->xres - 1;
	if (blk_prm->action(blk_prm) == -1) return (-1);

	blk_prm->x++;
	blk_prm->offset++;
	if (blk_prm->action(blk_prm) == -1) return (-1);

	return (0);
	}



/****************************************************************************
 * block_too_small
 ****************************************************************************/
static int
block_too_small(
	int				x,
	int				y)

	{
	if (x * y < 256) return (1);
	return (0);
	}



/****************************************************************************
 * block_msb
 ****************************************************************************/
static void
block_msb(
	struct codec			*cdc,
	struct block			*blk,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend)

	{
	int				i, x, y, m;

	*blk = BLOCK_INIT(xstart + xend, ystart + yend);
	for (y = ystart; y < yend; y++) for (i = y * cdc->xres + xstart, x = xstart; x < xend; i++, x++)
		{
		m = pcoeff_get_msb(&cdc->cft[i].pcf);
		if (m > blk->msb) blk->msb = m;
		}
	debug(1, "(%d,%d)-(%d,%d) msb = %d", xstart, ystart, xend, yend, blk->msb);
	}



/****************************************************************************
 * block_get_significance
 ****************************************************************************/
static int
block_get_significance(
	struct block			*blk,
	int				x,
	int				y,
	int				bitplane)

	{
	int				i;

	for (i = 0; blk[i].x + blk[i].y > 0; i++) if ((blk[i].x == x) && (blk[i].y == y)) break;
	debug_error_condition(blk[i].msb > bitplane);
	if (blk[i].msb == bitplane) return (1);
	return (0);
	}



/****************************************************************************
 * block_found
 ****************************************************************************/
static int
block_found(
	struct block			*blk,
	int				x,
	int				y)

	{
	int				i;

	for (i = 0; blk[i].x + blk[i].y > 0; i++) if ((blk[i].x == x) && (blk[i].y == y)) return (1);
	return (0);
	}



/****************************************************************************
 * block_append
 ****************************************************************************/
static void
block_append(
	struct block			*blk,
	int				x,
	int				y)

	{
	int				i;

	for (i = 0; blk[i].x + blk[i].y > 0; i++) ;
	blk[i] = BLOCK_INIT(x, y);
	}



/****************************************************************************
 * block_significance
 ****************************************************************************/
static int
block_significance(
	struct block_params		*blk_prm)

	{
	struct ac			*ac  = blk_prm->cdc->ac;
	struct block			*blk = blk_prm->blk_signif;
	int				x1   = blk_prm->xend - blk_prm->xstart;
	int				y1   = blk_prm->yend - blk_prm->ystart;
	int				x2   = blk_prm->xend + blk_prm->xstart;
	int				y2   = blk_prm->yend + blk_prm->ystart;
	int				bit;

	if (block_too_small(x1, y1)) return (1);
	if (block_found(blk_prm->blk_signif, x2, y2)) return (1);
	if (block_found(blk_prm->blk_insignif, x2, y2)) return (0);
	if (blk_prm->blk_msb != NULL)
		{
		bit = block_get_significance(blk_prm->blk_msb, x2, y2, blk_prm->bitplane);
		if (ac_ctx_write_bit(ac, CTX_BLOCK, bit) == -1) return (-1);
		}
	else
		{
		bit = ac_ctx_read_bit(ac, CTX_BLOCK);
		if (bit == -1) return (-1);
		}
	if (! bit) blk = blk_prm->blk_insignif;
	block_append(blk, x2, y2);
	return (bit);
	}



/****************************************************************************
 * iterate_block
 ****************************************************************************/
static int
iterate_block(
	struct block_params		*blk_prm)

	{
	int				xres   = blk_prm->cdc->xres;
	int				xres2  = xres / 2;
	int				result = block_significance(blk_prm);
	int				tupel[TUPEL_SIZE] = { };

	debug(1, "(%d,%d)-(%d,%d) result = %d", blk_prm->xstart, blk_prm->ystart, blk_prm->xend, blk_prm->yend, result);
	if (result != 1) return (result);
	blk_prm->tupel = tupel;
	if (((blk_prm->xstart % 2) + (blk_prm->xend % 2) + (blk_prm->ystart % 2) + (blk_prm->yend % 2) > 0) ||
		(blk_prm->xstart + blk_prm->ystart == 0))
		{
		blk_prm->offset2 = 0;
		for (blk_prm->y = blk_prm->ystart; blk_prm->y < blk_prm->yend; blk_prm->y++) for (blk_prm->offset = blk_prm->y * xres + blk_prm->xstart, blk_prm->x = blk_prm->xstart; blk_prm->x < blk_prm->xend; blk_prm->offset++, blk_prm->x++)
			{
			if (blk_prm->action(blk_prm) == -1) return (-1);
			}
		}
	else
		{
		for (blk_prm->y2 = blk_prm->ystart2; blk_prm->y2 < blk_prm->yend2; blk_prm->y2++) for (blk_prm->offset2 = blk_prm->y2 * xres2 + blk_prm->xstart2, blk_prm->x2 = blk_prm->xstart2; blk_prm->x2 < blk_prm->xend2; blk_prm->offset2++, blk_prm->x2++)
			{
			if (action2x2(blk_prm) == -1) return (-1);
			}
		}
	return (0);
	}



/****************************************************************************
 * iterate_fractional_bitplane2
 ****************************************************************************/
static int
iterate_fractional_bitplane2(
	struct block_params		*blk_prm)

	{
	int				x = blk_prm->cdc->xres >> blk_prm->cdc->level;
	int				y = blk_prm->cdc->yres >> blk_prm->cdc->level;
	int				l;

	if (iterate_block(set_coords(blk_prm, 0, 0, x, y, SUBBAND_LL)) == -1) return (-1);
	for (l = 0; l < blk_prm->cdc->level; l++, x <<= 1, y <<= 1)
		{
		if (iterate_block(set_coords(blk_prm, x, 0, x + x,     y, SUBBAND_LH)) == -1) return (-1);
		if (iterate_block(set_coords(blk_prm, 0, y,     x, y + y, SUBBAND_HL)) == -1) return (-1);
		if (iterate_block(set_coords(blk_prm, x, y, x + x, y + y, SUBBAND_HH)) == -1) return (-1);
		}
	return (0);
	}



/****************************************************************************
 * iterate_fractional_bitplane
 ****************************************************************************/
static int
iterate_fractional_bitplane(
	struct block_params		*blk_prm)

	{
	int				result = -1;

	blk_prm->cdc->signif = 0;
	blk_prm->cdc->sign   = 0;
	blk_prm->cdc->magn   = 0;
	if ((blk_prm->bitplane != blk_prm->cdc->bitplane) || (blk_prm->pass <= blk_prm->cdc->pass))
		{
		result = iterate_fractional_bitplane2(blk_prm);
		blk_prm->cdc->sum_signif += blk_prm->cdc->signif;
		blk_prm->cdc->sum_sign   += blk_prm->cdc->sign;
		blk_prm->cdc->sum_magn   += blk_prm->cdc->magn;
		if (blk_prm->pass > 0) verbose(2, "bitplane %2d:%d: signif =%9d/%-9d, sign =%9d/%-9d, magn =%9d/%-9d",
			blk_prm->bitplane, blk_prm->pass, blk_prm->cdc->signif, blk_prm->cdc->sum_signif,
			blk_prm->cdc->sign, blk_prm->cdc->sum_sign, blk_prm->cdc->magn,
			blk_prm->cdc->sum_magn);
		}
	return (result);
	}



/****************************************************************************
 * iterate_bitplane
 ****************************************************************************/
static int
iterate_bitplane(
	struct block_params		*blk_prm)

	{
	struct block			blk_insignif[64] = { };
	int				(*action_signif)(struct block_params *) = action_encode_significance;
	int				(*action_magn)(struct block_params *)   = action_encode_magnitude;
	int				ctx;

	blk_prm->blk_insignif = blk_insignif;
	if ((blk_prm->bitplane <= 3) || (blk_prm->cdc->sum_sign >= blk_prm->cdc->xres * blk_prm->cdc->yres / 8))
		{
		for (ctx = 0; ctx < CTXS_PER_SUBBAND; ctx++)
			{
			ac_ctx_reset(blk_prm->cdc->ac, CTX_BASE + ctx, 2, 1);
			ac_ctx_reset(blk_prm->cdc->ac, CTX_BASE + CTXS_PER_SUBBAND + ctx, 2, 1);
			ac_ctx_reset(blk_prm->cdc->ac, CTX_BASE + 2 * CTXS_PER_SUBBAND + ctx, 2, 1);
			ac_ctx_reset(blk_prm->cdc->ac, CTX_BASE + 3 * CTXS_PER_SUBBAND + ctx, 2, 1);
			}
		blk_prm->ignore_distance = 1;
		}
	if (blk_prm->blk_msb == NULL)
		{
		action_signif = action_decode_significance;
		action_magn   = action_decode_magnitude;
		}
	if (iterate_fractional_bitplane(set_pass(blk_prm, action_signif, select2x2_significance1, ctx_select1, 1)) == -1) return (-1);
	if (iterate_fractional_bitplane(set_pass(blk_prm, action_signif, select2x2_significance1, ctx_select2, 2)) == -1) return (-1);
	if (iterate_fractional_bitplane(set_pass(blk_prm, action_signif, select2x2_significance1, ctx_select3, 3)) == -1) return (-1);
	if (iterate_fractional_bitplane(set_pass(blk_prm, action_signif, select2x2_significance2, ctx_select4, 4)) == -1) return (-1);
	if (iterate_fractional_bitplane(set_pass(blk_prm, action_magn,   select2x2_magnitude,     NULL,        5)) == -1) return (-1);
	if (iterate_fractional_bitplane(set_pass(blk_prm, action_signif, select2x2_significance2, ctx_select5, 6)) == -1) return (-1);
	if (iterate_fractional_bitplane(set_pass(blk_prm, action_signif, select2x2_significance3, ctx_select6, 7)) == -1) return (-1);
	if (blk_prm->bitplane > 0) iterate_fractional_bitplane(set_pass(blk_prm, action_update, select2x2_update, NULL, -1));
	return (0);
	}



/****************************************************************************
 * iterate
 ****************************************************************************/
static void
iterate(
	struct codec			*cdc,
	struct block			*blk_msb)

	{
	struct block			blk_signif[64] = { };
	struct block_params		blk_prm = { .cdc = cdc, .blk_msb = blk_msb, .blk_signif = blk_signif };
	int				size = (cdc->xres * cdc->yres * sizeof (unsigned char)) / 4;
	int				n;

	blk_prm.signif2x2 = malloc(size);
	if (blk_prm.signif2x2 == NULL) error_oom();
	memset(blk_prm.signif2x2, 0, size);
	ac_init2(cdc->ac, CTX_BASE + 4 * CTXS_PER_SUBBAND, 1, 1);
	for (blk_prm.bitplane = cdc->msb, n = 0; (blk_prm.bitplane >= cdc->bitplane) && (n < cdc->nr_bitplanes); blk_prm.bitplane--, n++)
		{
		if (iterate_bitplane(&blk_prm) == -1) break;
		}
	free(blk_prm.signif2x2);
	}



/****************************************************************************
 * codec007_encode
 ****************************************************************************/
void
codec007_encode(
	struct codec			*cdc)

	{
	struct block			blk_msb[64] = { };
	int				x = cdc->xres;
	int				y = cdc->yres;
	int				i, l;

	debug_error_condition(cdc->level > 16);
	for (i = l = 0; l < cdc->level; l++)
		{
		x >>= 1;
		y >>= 1;
		block_msb(cdc, &blk_msb[i++], x, 0, x + x,     y);
		block_msb(cdc, &blk_msb[i++], 0, y,     x, y + y);
		block_msb(cdc, &blk_msb[i++], x, y, x + x, y + y);
		if (block_too_small(x >> 1, y >> 1)) break;
		}
	block_msb(cdc, &blk_msb[i], 0, 0, x, y);
	for (cdc->msb = 0; blk_msb[i].x + blk_msb[i].y > 0; i++)
		{
		if (blk_msb[i].msb > cdc->msb) cdc->msb = blk_msb[i].msb;
		}
	iterate(cdc, blk_msb);
	}



/****************************************************************************
 * codec007_decode
 ****************************************************************************/
void
codec007_decode(
	struct codec			*cdc)

	{
	debug_error_condition(cdc->level > 16);
	iterate(cdc, NULL);
	}
/******************************************************** Karsten Scheibler */
