/****************************************************************************
 ****************************************************************************
 *
 * pcoeff.c
 *
 ****************************************************************************
 ****************************************************************************/





#include "pcoeff.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "picture.h"



/****************************************************************************
 * get_msb
 ****************************************************************************/
static int
get_msb(
	unsigned int			val)

	{
	int				msb;

	for (msb = 0; val > 0; msb++) val >>= 1;
	if (msb > 0) msb--;
	return (msb);
	}



/****************************************************************************
 * pcoeff_set
 ****************************************************************************/
void
pcoeff_set(
	struct pcoeff			*pcf,
	double				val,
	int				lsb)

	{
	*pcf = (struct pcoeff) { };
	if (val < 0) val = -val, pcf->sign = 1;
	debug_error_condition(val > 4294967295.0);
	pcf->magn = (unsigned int) val;
	pcf->msb  = get_msb(pcf->magn);
	pcf->lsb  = lsb;
	}



/****************************************************************************
 * pcoeff_get
 ****************************************************************************/
double
pcoeff_get(
	struct pcoeff			*pcf)

	{
	double				val;

	if (pcf->magn == 0) return (0.0);
	val = ((double) pcf->magn) + 0.5 * ((double) (1 << pcf->valid));
	if (pcf->sign) val = -val;
	return (val);
	}



/****************************************************************************
 * pcoeff_already_finished
 ****************************************************************************/
int
pcoeff_already_finished(
	struct pcoeff			*pcf,
	int				bitplane)

	{
	return ((pcf->lsb > bitplane) ? 1 : 0);
	}



/****************************************************************************
 * pcoeff_already_significant
 ****************************************************************************/
int
pcoeff_already_significant(
	struct pcoeff			*pcf,
	int				bitplane)

	{

	/*
	 * bitplane is needed to return correct results while encoding and
	 * decoding
	 */

	debug_error_condition(bitplane < pcf->lsb);
	if (pcf->magn == 0) return (0);
	bitplane -= pcf->msb + pcf->lsb;
	return ((bitplane < 0) ? 1 : 0);
	}



/****************************************************************************
 * pcoeff_get_sign
 ****************************************************************************/
int
pcoeff_get_sign(
	struct pcoeff			*pcf)

	{
	return (pcf->sign);
	}



/****************************************************************************
 * pcoeff_get_msb
 ****************************************************************************/
int
pcoeff_get_msb(
	struct pcoeff			*pcf)

	{
	return (pcf->msb + pcf->lsb);
	}



/****************************************************************************
 * pcoeff_get_significance
 ****************************************************************************/
int
pcoeff_get_significance(
	struct pcoeff			*pcf,
	int				bitplane)

	{
	debug_error_condition(bitplane < pcf->lsb);
	if (pcf->magn == 0) return (0);
	bitplane -= pcf->msb + pcf->lsb;
	return ((bitplane <= 0) ? 1 : 0);
	}



/****************************************************************************
 * pcoeff_get_magnitude
 ****************************************************************************/
int
pcoeff_get_magnitude(
	struct pcoeff			*pcf,
	int				bitplane)

	{
	debug_error_condition(bitplane < pcf->lsb);
	bitplane -= pcf->lsb;
	debug_error_condition(pcf->msb <= bitplane);
	return ((pcf->magn & (1 << bitplane)) ? 1 : 0);
	}



/****************************************************************************
 * pcoeff_set_significance_and_sign
 ****************************************************************************/
void
pcoeff_set_significance_and_sign(
	struct pcoeff			*pcf,
	int				bitplane,
	int				sign)

	{
	debug_error_condition(bitplane < pcf->lsb);
	bitplane   -= pcf->lsb;
	pcf->msb   = bitplane;
	pcf->valid = bitplane;
	pcf->magn  = 1 << bitplane;
	pcf->sign  = sign;
	}



/****************************************************************************
 * pcoeff_set_magnitude
 ****************************************************************************/
void
pcoeff_set_magnitude(
	struct pcoeff			*pcf,
	int				bitplane,
	int				bit)

	{
	debug_error_condition((bit != 0) && (bit != 1));
	debug_error_condition(bitplane < pcf->lsb);
	bitplane -= pcf->lsb;
	debug_error_condition(pcf->msb <= bitplane);
	if (bit) pcf->magn |= 1 << bitplane;
	pcf->valid = bitplane;
	}



/****************************************************************************
 * pcoeff_ctx_get1
 ****************************************************************************/
int
pcoeff_ctx_get1(
	struct pcoeff			*pcf)

	{
	if (pcf->ctx > 7) return (7);
	return (pcf->ctx);
	}



/****************************************************************************
 * pcoeff_ctx_get2
 ****************************************************************************/
int
pcoeff_ctx_get2(
	struct pcoeff			*pcf)

	{
	return (pcf->ctx);
	}



/****************************************************************************
 * pcoeff_ctx_already_done
 ****************************************************************************/
int
pcoeff_ctx_already_done(
	struct pcoeff			*pcf)

	{
	return (pcf->done);
	}



/****************************************************************************
 * pcoeff_ctx_update
 ****************************************************************************/
void
pcoeff_ctx_update(
	struct pcoeff			*pcf)

	{
	pcf->ctx  = (pcf->ctx < 8) ? (pcf->ctx << 1) : 15;
	pcf->done = 0;
	}



/****************************************************************************
 * pcoeff_ctx_set_significance
 ****************************************************************************/
void
pcoeff_ctx_set_significance(
	struct pcoeff			*pcf)

	{
	pcf->ctx = 1;
	}



/****************************************************************************
 * pcoeff_ctx_set_magnitude
 ****************************************************************************/
void
pcoeff_ctx_set_magnitude(
	struct pcoeff			*pcf,
	int				bit)

	{
	debug_error_condition((bit != 0) && (bit != 1));
	pcf->ctx |= bit;
	}



/****************************************************************************
 * pcoeff_ctx_set_done
 ****************************************************************************/
void
pcoeff_ctx_set_done(
	struct pcoeff			*pcf)

	{
	pcf->done = 1;
	}



/****************************************************************************
 * pcoeff_dist_get
 ****************************************************************************/
int
pcoeff_dist_get(
	struct pcoeff			*pcf)

	{
	return (pcf->dist);
	}



/****************************************************************************
 * pcoeff_dist_set
 ****************************************************************************/
void
pcoeff_dist_set(
	struct pcoeff			*pcf,
	int				dist)

	{
	if (dist > pcf->dist) pcf->dist = dist;
	}
/******************************************************** Karsten Scheibler */
