/****************************************************************************
 ****************************************************************************
 *
 * pgm.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdio.h>
#include <stdlib.h>

#include "pgm.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "picture.h"



/****************************************************************************
 * write_info
 ****************************************************************************/
static void
write_info(
	char				*action,
	char				*path,
	int				xres,
	int				yres)

	{
	verbose(1, "%s PGM file '%s', resolution: %dx%d", action, path, xres, yres);
	}



/****************************************************************************
 * read_version2
 ****************************************************************************/
static void
read_version2(
	FILE				*fh,
	unsigned char			*data,
	int				size)

	{
	int				i, v;

	for (i = 0; i < size; i++)
		{
		fscanf(fh, "%d", &v);
		data[i] = v;
		}
	}



/****************************************************************************
 * pgm_read
 ****************************************************************************/
void
pgm_read(
	char				*path,
	struct picture			*pic)

	{
	FILE				*fh;
	char				version;
	int				xres, yres;
	unsigned char			*data;

	fh = fopen(path, "r");
	if (fh == NULL) error_perror2("error while opening '%s'", path);

	/* read file header */

		{
		int			comment;
		char			magic;
		char			line[1024];
		int			grey_levels;

		fscanf(fh, "%c", &magic);
		fscanf(fh, "%c", &version);
		fgetc(fh);
		if ((magic != 'P') || ((version != '5') && (version != '2'))) error("'%s' is not a pgm file", path);

		while (1)
			{
			comment = fgetc(fh);
			if (comment != '#') break;
			fgets(line, sizeof (line), fh);
			}
		ungetc(comment, fh);

		fscanf(fh, "%d", &xres);
		fscanf(fh, "%d", &yres);
		fscanf(fh, "%d", &grey_levels);
		fgetc(fh);
		if (grey_levels != 255) error("file '%s' does not have 255 grey levels", path);
		}
	write_info("reading", path, xres, yres);

	/* allocate memory */

	picture_init(pic, xres, yres);
	data = (unsigned char *) malloc(xres * yres * sizeof (unsigned char));
	if (data == NULL) error_oom();

	/* read file data */

	if (version == '2') read_version2(fh, data, xres * yres);
	else fread(data, yres * xres, 1, fh);
	picture_write_uchar(pic, data);
	free(data);
	fclose(fh);
	}



/****************************************************************************
 * pgm_write
 ****************************************************************************/
void
pgm_write(
	char				*path,
	struct picture			*pic)

	{
	FILE				*fh;
	unsigned char			*data;
	int				xres = picture_get_xres(pic);
	int				yres = picture_get_yres(pic);

	write_info("writing", path, xres, yres);
	data = (unsigned char *) malloc(xres * yres * sizeof (unsigned char));
	if (data == NULL) error_oom();
	picture_read_uchar(pic, data);
	fh = fopen(path, "w");
	if (fh == NULL) error_perror2("error while opening '%s'", path);
	fprintf(fh, "P5\n# generated by yawica\n%d %d\n255\n", xres, yres);
	fwrite(data, xres * yres, 1, fh);
	free(data);
	fclose(fh);
	}
/******************************************************** Karsten Scheibler */
