/****************************************************************************
 ****************************************************************************
 *
 * picture.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdlib.h>
#include <string.h>

#include "picture.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"



/****************************************************************************
 * cap
 ****************************************************************************/
static unsigned char
cap(
	double				val)

	{
	if (val < 0.0) val = 0.0;
	if (val > 255.0) val = 255.0;
	return ((unsigned char) val);
	}



/****************************************************************************
 * picture_init
 ****************************************************************************/
void
picture_init(
	struct picture			*pic,
	int				xres,
	int				yres)

	{
	debug(1, "allocating picture");
	*pic = (struct picture)
		{
		.cft  = (union coefficient *) malloc(xres * yres * sizeof (union coefficient)),
		.mode = PICTURE_MODE_DOUBLE,
		.xres = xres,
		.yres = yres
		};
	if (pic->cft == NULL) error_oom();
	}



/****************************************************************************
 * picture_deinit
 ****************************************************************************/
void
picture_deinit(
	struct picture			*pic)

	{
	debug(1, "freeing picture");
	debug_error_condition(pic->cft == NULL);
	free(pic->cft);
	}



/****************************************************************************
 * picture_write_uchar
 ****************************************************************************/
void
picture_write_uchar(
	struct picture			*pic,
	unsigned char			*data)

	{
	int				i, j = pic->xres * pic->yres;

	debug(1, "writing %s to picture", (data == NULL) ? "0" : "content");
	pic->mode = PICTURE_MODE_DOUBLE;
	if (data == NULL) for (i = 0; i < j; i++) pic->cft[i].d = 0.0;
	else for (i = 0; i < j; i++) pic->cft[i].d = ((double) data[i]) - 128.0;
	}



/****************************************************************************
 * picture_read_uchar
 ****************************************************************************/
void
picture_read_uchar(
	struct picture			*pic,
	unsigned char			*data)

	{
	int				i, j = pic->xres * pic->yres;

	debug(1, "read from picture");
	debug_error_condition(data == NULL);
	debug_error_condition(pic->mode != PICTURE_MODE_DOUBLE);
	for (i = 0; i < j; i++) data[i] = cap(pic->cft[i].d + 128.5);
	}



/****************************************************************************
 * picture_get_coefficient
 ****************************************************************************/
union coefficient *
picture_get_coefficient(
	struct picture			*pic,
	int				mode)

	{
	debug_error_condition(mode != pic->mode);
	debug_error_condition(pic->cft == NULL);
	return (pic->cft);
	}



/****************************************************************************
 * picture_get_xres
 ****************************************************************************/
int
picture_get_xres(
	struct picture			*pic)

	{
	return (pic->xres);
	}



/****************************************************************************
 * picture_get_yres
 ****************************************************************************/
int
picture_get_yres(
	struct picture			*pic)

	{
	return (pic->yres);
	}



/****************************************************************************
 * picture_get_shift
 ****************************************************************************/
int
picture_get_shift(
	struct picture			*pic)

	{
	return (pic->shift);
	}



/****************************************************************************
 * picture_set_mode
 ****************************************************************************/
void
picture_set_mode(
	struct picture			*pic,
	int				mode)

	{
	pic->mode = mode;
	}



/****************************************************************************
 * picture_set_shift
 ****************************************************************************/
void
picture_set_shift(
	struct picture			*pic,
	int				shift)

	{
	pic->shift = shift;
	}
/******************************************************** Karsten Scheibler */
