/****************************************************************************
 ****************************************************************************
 *
 * psnr.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <math.h>
#include <stdio.h>

#include "psnr.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "picture.h"



/****************************************************************************
 * psnr_calculate
 ****************************************************************************/
void
psnr_calculate(
	struct picture			*pic1,
	struct picture			*pic2,
	struct psnr			*psn)

	{
	union coefficient		*cft1, *cft2;
	int				i, size;
	double				val;

	if ((picture_get_xres(pic1) != picture_get_xres(pic2)) ||
		(picture_get_yres(pic1) != picture_get_yres(pic2))) error_message("picture resolutions are different");
	size = picture_get_xres(pic1) * picture_get_yres(pic1);
	cft1 = picture_get_coefficient(pic1, PICTURE_MODE_DOUBLE);
	cft2 = picture_get_coefficient(pic2, PICTURE_MODE_DOUBLE);
	for (psn->mse_n = i = 0; i < size; i++)
		{
		val = cft1[i].d - cft2[i].d;
		psn->mse_n += val * val;
		}
	psn->mse  = psn->mse_n / ((double) size);
	psn->psnr = 10 * log10((255 * 255) / psn->mse);
	}



/****************************************************************************
 * psnr_print
 ****************************************************************************/
void
psnr_print(
	struct psnr			*psn)

	{
	printf("MSE * N: %1.0f\n", psn->mse_n);
	printf("MSE:     %f\n", psn->mse);
	printf("PSNR:    %f\n", psn->psnr);
	}
/******************************************************** Karsten Scheibler */
